<?php

namespace Drupal\elasticsearch_connector_suggester\EventSubscriber;

use Drupal\Core\Config\ConfigEvents;
use Drupal\Core\Entity\EntityInterface;
use Drupal\elasticsearch_connector\ElasticSearch\Parameters\Factory\IndexFactory;
use Drupal\Component\EventDispatcher\Event;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\elasticsearch_connector\Event\BuildIndexParamsEvent;
use Drupal\elasticsearch_connector\Event\BuildSearchParamsEvent;
use Drupal\elasticsearch_connector\Event\PrepareIndexMappingEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Class DefaultSubscriber for elasticsearch connector suggester.
 */
class DefaultSubscriber implements EventSubscriberInterface {

  /**
   * Entity type manager.
   *
   * @var \Drupal\Core\Entity\Entity
   */
  private $entityTypeManager;

  /**
   * Constructs a new DefaultSubscriber object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_typeManager
   *   Entity type manager interface.
   */
  public function __construct(EntityTypeManagerInterface $entity_typeManager) {
    $this->entityTypeManager = $entity_typeManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    $events[PrepareIndexMappingEvent::PREPARE_INDEX_MAPPING] = [
      'onElasticsearchConnectorPrepareIndexMapping'];
    $events[BuildIndexParamsEvent::BUILD_PARAMS] = [
      'onElasticsearchBuildParams'];
    $events[BuildSearchParamsEvent::BUILD_QUERY] = [
      'onElasticsearchBuildQuery'];
    $events[ConfigEvents::SAVE] = ['onConfigSave'];

    return $events;
  }

  /**
   * Called on elasticsearch_connector.prepare_index_mapping event.
   * Add suggester field to index mapping.
   *
   * @param \Drupal\elasticsearch_connector\Event\PrepareIndexMappingEvent $event
   *   The event.
   */
  public function onElasticsearchConnectorPrepareIndexMapping(Event $event) {
    $fields = $this->getFields($event);
    if (empty($fields)) {
      return;
    }
    $params = $event->getIndexMappingParams();
    foreach ($fields as $field_id) {
      // Set suggester field.
      $params['body']['properties'][$field_id . '_suggest'] = [
        'type' => 'completion',
        'analyzer'=> 'standard',
        'search_analyzer'=> 'standard',
        'preserve_separators'=> true,
        'preserve_position_increments'=> true,
      ];
    }

    $event->setIndexMappingParams($params);
  }

  /**
   * Called on elasticsearch_connector.build_params event.
   * Add suggester value to index element.
   *
   * @param \Drupal\elasticsearch_connector\Event\BuildIndexParamsEvent $event
   *   The event.
   */
  public function onElasticsearchBuildParams(Event $event) {
    $fields = $this->getFields($event);
    if (empty($fields)) {
      return;
    }

    $params = $event->getElasticIndexParams();
    foreach ($fields as $field) {
      if (!isset($params['body'][1][$field])) {
        continue;
      }
      $suggested_name = $field . '_suggest';
      // Set suggester value.
      $params['body'][1][$suggested_name]['input'] = $params['body'][1][$field];
    }

    $event->setElasticIndexParams($params);
  }

  /**
   * Called on elasticsearch_connector.prepare_query event.
   * Replace query to suggester request.
   *
   * @param \Drupal\elasticsearch_connector\Event\BuildSearchParamsEvent $event
   *   The event.
   */
  public function onElasticsearchBuildQuery(Event $event) {
    $params = $event->getElasticSearchParams();
    $field_name = $params['body']['query']['bool']['must']['query_string']['fields'][0] ?? '';
    if (empty($field_name)) {
      return;
    }
    $query_string = $params['body']['query']['bool']['must']['query_string']['query'] ?? '';
    if (empty($query_string)) {
      return;
    }
    $string = str_replace('AND ', '', $query_string);
    if (empty($string)) {
      return;
    }

    // Prepare suggester field name.
    $name = $field_name;
    if (strpos($field_name, '_suggest') === FALSE) {
      // Remove from name the ^0 or same string replace with _suggest string.
      $name = preg_replace('/\^.*$/', '_suggest', $field_name);
    }
    if ($name === $field_name && strpos($name, '_suggest') === FALSE) {
      // Field name not contains _suggest string.
      $name .= '_suggest';
    }

    // Set suggest request.
    $params['body'] = [
      'suggest' => [
        'autocomplete' => [
          'prefix' => $string,
          'completion' => [
            'field' => $name,
            'size' => 10,
          ],
        ],
      ],
    ];

    $event->setElasticSearchParams($params);
  }

  /**
   * Called on config.save event.
   * Update elasticsearch index mapping.
   *
   * @param \Drupal\Core\Config\ConfigEvents $event
   *   The event.
   */
  public function onConfigSave(Event $event) {
    /** @var \Drupal\Core\Config\Config $config */
    $config = $event->getConfig();
    $name = $config->getName();
    if (!preg_match('/^search_api_autocomplete.search.*$/', $name)) {
      return;
    }
    // Save index to update mapping.
    if ($index = $this->entityTypeManager->getStorage('search_api_index')
      ->load($config->get('index_id'))) {
      $index->save();
    }
  }

  /**
   * Get autocomplete fields.
   *
   * @param \Drupal\Component\EventDispatcher\Event $event
   *   The event.
   *
   * @return array
   *   The fields.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getFields(Event $event): array {
    $fields = [];

    // Load all autocomplete entities.
    $autocompletes = $this->entityTypeManager
      ->getStorage('search_api_autocomplete_search')
      ->loadMultiple();
    $index = $this->loadIndexFromIndexName($event->getIndexName());

    /** @var \Drupal\search_api_autocomplete\Entity\Search $value */
    foreach ($autocompletes as $value) {
      // Get autocomplete settings for the index.
      if ($value->getIndexId() == $index->id()) {
        /** @var \Drupal\search_api_autocomplete\Entity\Search $autocomplete */
        $autocomplete = $value;
        break;
      }
    }

    if (!isset($autocomplete)) {
      // No autocomplete settings for the index.
      return $fields;
    }

    /** @var \Drupal\search_api_autocomplete\Suggester\SuggesterInterface $suggester */
    foreach ($autocomplete->getSuggesters() as $suggester) {
      // Get fields for the suggester.
      $fields[] = $suggester->getConfiguration()['fields'] ?? [];
    }

    return array_merge(...$fields);
  }

  /**
   * Loads the index entity associated with this event.
   *
   * @param string $index_name
   *   The long index name as a string.
   *
   * @return \Drupal\Core\Entity\EntityInterface|null
   *   The loaded index or NULL.
   */
  private function loadIndexFromIndexName($index_name):? EntityInterface {
    /** @var \Drupal\search_api\Entity\Index[] $search_api_indexes */
    $search_api_indexes = $this->entityTypeManager
      ->getStorage('search_api_index')
      ->loadMultiple();

    foreach ($search_api_indexes as $search_api_index) {
      $elasticsearch_connector_index_name = IndexFactory::getIndexName($search_api_index);
      if ($index_name == $elasticsearch_connector_index_name) {
        return $search_api_index;
      }
    }

    return NULL;
  }

}
